# AUTOGENERATED! DO NOT EDIT! File to edit: nbks/03_train.ipynb (unless otherwise specified).

__all__ = ['timm_or_fastai_arch', 'train', 'softmax_RocAuc', 'train_cv']

# Cell
from .dataset import *
from .evaluate import *

from fastai.vision.all import *
from fastcore.script import *
from fastai.callback.wandb import *
from wwf.vision.timm import *
import timm
import wandb
from typing import *
from sys import exit

# Cell
def timm_or_fastai_arch(arch: str) -> (Union[Any, str], Callable[..., Learner]):
    try:  # Check if fastai arch
        model = globals()[arch]
        learner_func = cnn_learner
    except KeyError:  # Must be timm arch
        model = arch
        learner_func = timm_learner
    return model, learner_func

# Cell
def train(
    data_path: Path, epochs: int = 1, lr: Union[float, str] = 3e-4, frz: int=1, pre: int=800, re: int=256,
    bs: int=200, fold: int=4, smooth: bool=False,
    arch: str='resnet18', dump: bool=False, log: bool=False, mixup: float=0.,
    fp16: bool=False, dls: DataLoaders=None, save: bool=False, pseudo: Path=None,
 ) -> Learner:
    # Prep Data, Opt, Loss, Arch
    if dls is None:
        dls = get_dls_all_in_1(
            data_path=data_path, presize=pre, resize=re, bs=bs, val_fold=fold, pseudo_labels_path=pseudo
        )
    if log: wandb.init(project="plant-pathology")
    if smooth: loss_func = LabelSmoothingCrossEntropyFlat()
    else:      loss_func = CrossEntropyLossFlat()
    m, learner_func = timm_or_fastai_arch(arch)

    # Add callbacks
    cbs = [SaveModelCallback("roc_auc_score", fname=f"model_val_on_{fold}")] if save or log else []
    if log: cbs.append(WandbCallback())
    if mixup: cbs.append(MixUp(mixup))

    # Build learner
    print(f"# train exs: {len(dls.train_ds)}, val exs: {len(dls.valid_ds)}")
    learn = learner_func(dls, m, loss_func=loss_func,
                    metrics=[accuracy, RocAuc()], cbs=cbs)
    if dump: print(learn.model); exit()
    if lr=="find": learn.lr_find(); exit()
    if fp16: learn.to_fp16()

    # Train
    learn.freeze()
    learn.fit_one_cycle(frz, lr)
    learn.unfreeze()
    learn.fit_one_cycle(epochs, slice(lr/100, lr/2))  # Explore other divs
    return learn

# Cell
def softmax_RocAuc(logits, labels):
    probs = logits.softmax(-1)
    return RocAuc()(probs, labels)

# Cell
@call_parse
def train_cv(
    path:     Param("Path to data dir", Path),
    epochs:   Param("Number of unfrozen epochs", int)=1,
    lr:       Param("Initial learning rate", float)=3e-4,
    frz:      Param("Number of frozen epochs", int)=1,
    pre:      Param("Image presize", int, nargs="+")=(682, 1024),
    re:       Param("Image resize", int)=256,
    bs:       Param("Batch size", int)=256,
    smooth:   Param("Label smoothing?", store_true)=False,
    arch:     Param("Architecture", str)='resnet18',
    dump:     Param("Don't train, just print model", store_true)=False,
    log:      Param("Log w/ W&B", store_true)=False,
    save:     Param("Save model based on RocAuc", store_true)=False,
    mixup:    Param("Mixup (0.4 is good)", float)=0.0,
    tta:      Param("Test-time augmentation", store_true)=False,
    fp16:     Param("Mixed-precision training", store_true)=False,
    do_eval: Param("Evaluate model and save predictions CSV", store_true)=False,
    val_fold: Param("Don't go cross-validation, just do 1 fold (or pass 9 "
                    "to train on all data)", int)=None,
    pseudo:   Param("Path to pseudo labels to train on", Path)=None,
    export:   Param("Export learner(s) to export_val_on_{fold}.pkl", store_true)=False,
):
    print(locals())
    scores = []
    for fold in range(5):
        if val_fold is not None: fold = val_fold  # Not doing CV
        print(f"\nTraining on fold {fold}")
        learn = train(data_path=path, epochs=epochs, lr=lr, frz=frz, pre=pre,
                      re=re, bs=bs, smooth=smooth, arch=arch, dump=dump, log=log,
                      fold=fold, mixup=mixup, fp16=fp16, save=save, pseudo=pseudo)

        if hasattr(learn, "mixup") and tta: learn.remove_cb(MixUp)  # Bug when doing tta w/Mixup

        if tta and val_fold != 9:  # There IS a valid set
            preds, lbls = learn.tta()
            res = [f(preds, lbls) for f in [learn.loss_func, accuracy, softmax_RocAuc]]
        else: res = learn.final_record
        scores.append(res)

        # Create submission file for this model
        if do_eval: print("Evaluating"); evaluate(learn, path=path/"test.csv", name=f"predictions_fold_{fold}.csv", tta=tta)

        if export: learn.export(f"export_val_on_{fold}.pkl")
        # Delete learner to avoid OOM
        del learn
        if val_fold is not None: break
    scores = np.array(scores)
    print(f"Scores: {scores}\n")
    if val_fold is None: print(f"Mean: {scores.mean(0)}")


# AUTOGENERATED! DO NOT EDIT! File to edit: nbks/01_dataset.ipynb (unless otherwise specified).

__all__ = ['get_datablock', 'get_dls', 'get_dls_all_in_1']

# Cell
from fastai.vision.all import *
from .utils import *
from typing import *
import pandas as pd

# Cell
def _get_y(row, df: pd.DataFrame):
    return df.columns[row == 1][0]

# Cell
def get_datablock(path: Path, df: pd.DataFrame, presize: int,
                  resize: int, val_fold: int=4) -> DataBlock:
    return DataBlock(blocks=(ImageBlock, CategoryBlock),
                get_x=ColReader("image_id", pref=path/'images', suff=".jpg"),
                get_y=partial(_get_y, df=df),
                splitter=MaskSplitter(df["fold"]==val_fold),
                item_tfms=Resize(presize),
                batch_tfms=aug_transforms(mult=1.5, max_rotate=22.5, min_zoom=0.9,
                                         size=resize, min_scale=0.5, flip_vert=True,
                                         max_zoom=1.2))

# Cell
def get_dls(path: Path, df: pd.DataFrame, presize: Union[tuple, int]=(682, 1024),
            resize: int=256, bs: int=256, val_fold: int=4) -> DataLoaders:
    return get_datablock(path, df, presize, resize, val_fold).dataloaders(df, bs=bs)

# Cell
@delegates(get_dls)
def get_dls_all_in_1(data_path: Path, pseudo_labels_path: str=None, **kwargs) -> DataLoaders:
    path, df = load_data(data_path, with_folds=True, pseudo_labels_path=pseudo_labels_path)
    return get_dls(path, df, **kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbks/07_pretrained_models.ipynb (unless otherwise specified).

__all__ = ['MODELS', 'get_model']

# Cell
from fastai.data.external import untar_data
from fastai.learner import load_learner
from fastcore.test import ExceptionExpected

# Cell
MODELS = {
    "resnet18_2021-04-08": "https://github.com/bwolfson97/plant_pathology/releases/download/v0.1.1-alpha/resnet18_2021-04-08.tar.gz"
}

# Cell
def get_model(model_name: str):
    """Downloads and builds pretrained model."""
    try:
        url = MODELS[model_name]
    except KeyError:
        raise KeyError(f"Invalid model name. Received: {model_name}")

    pickle_file = untar_data(url)
    return load_learner(pickle_file)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbks/04_generate_pseudo_labels.ipynb (unless otherwise specified).

__all__ = ['generate_pseudo_labels']

# Cell
from fastcore.all import *
import pandas as pd
import numpy as np
from .utils import *

# Cell
@call_parse
def generate_pseudo_labels(
    path:   Param("Directory of prediction CSVs to average", Path)=".",
    name:   Param("File name to save as", str)="pseudo_labels.csv",
    thresh: Param("Min probabilty for pseudo label", float)=0.95,
) -> Path:
    """Generates pseudo labels and saves in path dir."""
    avg_preds_df = get_averaged_preds(path)

    # One-hot encode highly confident predictions
    high_confidence_preds_mask = avg_preds_df >= thresh
    avg_preds_df[high_confidence_preds_mask] = 1.0
    avg_preds_df[~high_confidence_preds_mask] = 0.0

    # Only keep predictions model was highly confident on
    pseudo_labels = avg_preds_df[high_confidence_preds_mask.any(axis=1)]

    pseudo_labels.to_csv(path/name)
    return path/name


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"load_data": "00_utils.ipynb",
         "kaggle_submit_command": "00_utils.ipynb",
         "average_preds": "00_utils.ipynb",
         "get_averaged_preds": "00_utils.ipynb",
         "get_datablock": "01_dataset.ipynb",
         "get_dls": "01_dataset.ipynb",
         "get_dls_all_in_1": "01_dataset.ipynb",
         "infer_on_test_set": "02_evaluate.ipynb",
         "format_submission": "02_evaluate.ipynb",
         "evaluate": "02_evaluate.ipynb",
         "timm_or_fastai_arch": "03_train.ipynb",
         "train": "03_train.ipynb",
         "softmax_RocAuc": "03_train.ipynb",
         "train_cv": "03_train.ipynb",
         "generate_pseudo_labels": "04_generate_pseudo_labels.ipynb",
         "create_folds": "06_create_folds.ipynb",
         "MODELS": "07_pretrained_models.ipynb",
         "get_model": "07_pretrained_models.ipynb",
         "DATA_PATH": "config.ipynb",
         "TEST_DATA_PATH": "config.ipynb"}

modules = ["utils.py",
           "dataset.py",
           "evaluate.py",
           "train.py",
           "generate_pseudo_labels.py",
           "self_knowledge_distillation.py",
           "create_folds.py",
           "pretrained_models.py",
           "config.py"]

doc_url = "https://bwolfson97.github.io/plant_pathology/"

git_url = "https://github.com/bwolfson97/plant_pathology/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbks/00_utils.ipynb (unless otherwise specified).

__all__ = ['load_data', 'kaggle_submit_command', 'average_preds', 'get_averaged_preds']

# Cell
from fastcore.all import *
from typing import Tuple, List
import pandas as pd
from sklearn.model_selection import StratifiedKFold

# Cell
def load_data(data_path: Path, with_folds: bool = False, pseudo_labels_path: str = None) -> Tuple[Path, pd.DataFrame]:
    """Load data (with/without cross-validation folds) into DataFrame."""
    train_df = pd.read_csv(data_path/('train_folds.csv' if with_folds else 'train.csv'))
    if pseudo_labels_path is not None:
        # Add pseudo labels to DataFrame
        train_df = pd.concat([train_df, pd.read_csv(pseudo_labels_path)], ignore_index=True)
    return data_path, train_df

# Cell
def kaggle_submit_command() -> str:
    """Print terminal command to submit submission file."""
    print("kaggle competitions submit -c plant-pathology-2020-fgvc7 -f {submission_path} -m 'message'")

# Cell
def average_preds(dfs: List[pd.DataFrame]) -> pd.DataFrame:
    """Average predictions on test examples across prediction DataFrames in `dfs`."""
    all_preds_df = pd.concat(dfs)
    avg_preds_df = all_preds_df.groupby(all_preds_df.image_id).mean()
    return avg_preds_df

# Cell
def get_averaged_preds(path: Path, verbose: bool = False) -> Path:
    """Returns DataFrame of averaged of averaged predictions of prediction CSVs in `path` dir."""
    # Load test set prediction CSVs for each of 5 CV folds
    prediction_files = list(path.glob("predictions_fold_[0-4].csv"))
    if verbose:
        print(prediction_files)
    return average_preds([pd.read_csv(fn) for fn in prediction_files])


# AUTOGENERATED! DO NOT EDIT! File to edit: nbks/02_evaluate.ipynb (unless otherwise specified).

__all__ = ['infer_on_test_set', 'format_submission', 'evaluate']

# Cell
from .config import TEST_DATA_PATH
from fastai.vision.all import *
from typing import *

# Cell
def infer_on_test_set(
    learn: Learner, path: Path, tta: bool = False, bs: int = 64,
) -> Tensor:
    """Infers on test CSV at `path` using `learn`, optionally performing TTA."""
    df_test = pd.read_csv(path)
    test_dl = learn.dls.test_dl(df_test, bs=bs)
    preds, _ = (learn.tta if tta else learn.get_preds)(dl=test_dl)
    return preds

# Cell
def format_submission(preds: Tensor, save_path: Union[Path, str]) -> Path:
    # Build submission CSV
    image_filenames = [f"Test_{i}" for i in range(len(preds))]
    column_names = ["healthy", "multiple_diseaes", "rust", "scab"]
    submission = pd.DataFrame(preds, index=image_filenames, columns=column_names)

    # Make parent dirs
    save_path = Path(save_path)
    Path(save_path.parent).mkdir(parents=True, exist_ok=True)

    # Save submission
    submission.to_csv(save_path)
    return save_path

# Cell
def evaluate(learn: Learner, path: Path, name: str = "submission.csv", tta: bool=False) -> Path:
    """Evaluates `learn` on test CSV at `path` and saves as `name`, optionally applying TTA."""
    preds = infer_on_test_set(learn, path=path, tta=tta)
    return format_submission(preds, name)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbks/config.ipynb (unless otherwise specified).

__all__ = ['DATA_PATH', 'TEST_DATA_PATH']

# Cell
from pathlib import Path

# Cell
# Change to reflect your machine
DATA_PATH = Path("../data/")
TEST_DATA_PATH = Path("../test_data/")


# AUTOGENERATED! DO NOT EDIT! File to edit: nbks/06_create_folds.ipynb (unless otherwise specified).

__all__ = ['create_folds']

# Cell
from pathlib import Path
import pandas as pd
from sklearn.model_selection import StratifiedKFold
from fastcore.script import call_parse, Param
from fastcore.script import store_true, bool_arg

# Cell
@call_parse
def create_folds(
    path:        Param("Path to train CSV", Path),
    print_stats: Param("Print class distributions across folds?", store_true),
    save: Param("Save CSV with added folds", bool_arg)=True,
) -> pd.DataFrame:
    """Saves train CSV at `path` with 5-fold CV splits added.

    Optionally, print class distribution across folds.
    """
    # Load train CSV
    df = pd.read_csv(path)

    # Shuffle rows
    df = df.sample(frac=1.0, random_state=42).reset_index(drop=True)

    # Get class labels to stratify on
    lbls = df.apply(lambda r: df.columns[r==1].item(), axis=1)

    # Create 5 folds
    kf = StratifiedKFold(n_splits=5)
    for fold, (_, val_idxs) in enumerate(kf.split(df, lbls.values)):
        df.loc[val_idxs, "fold"] = fold

    # Correct dtype of folds column
    df.fold = df.fold.astype(int)

    if print_stats:
        stats_df = df.groupby("fold").describe()
        print("Proportion of each class out of total examples in each fold:")
        print(stats_df.iloc[:, stats_df.columns.get_level_values(1) == "mean"])

    if save:
        save_path = path.parent/"train_folds.csv"
        df.to_csv(save_path, index=False)
        print(f"Saved to {save_path}")

    return df